"""
Zork: The Haunted Mansion Adventure RPG
A recreated thematic adventure game based on the original 1977 release of Zork I
with newly added turn-based RPG elements and a horror theme. Originally written 
in Python 2.5, but syntactically updated to Python 3 and is compatible with v3.6. 
Created and Developed by Chad Jordan
6/10/2007 updated 8/20/2017
"""
from dataclasses import dataclass
import random
import sys
import time
import builtins
import shutil

try:
    terminal_width = shutil.get_terminal_size().columns
except:
    terminal_width = 80

original_print = builtins.print
original_input = builtins.input

#Define a new print function with padding
def print_with_padding(*args, **kwargs):
    sep = kwargs.get('sep', ' ')
    end = kwargs.get('end', '\n')
    kwargs.pop('sep', None)
    kwargs.pop('end', None)
    
    #Convert all arguments to strings and join them
    text = sep.join(str(arg) for arg in args)
    lines = text.split('\n')
    
    for i, line in enumerate(lines):
        if i < len(lines) - 1:
            original_print(' ' * 5 + line, **kwargs)
        else:
            original_print(' ' * 5 + line, end=end, **kwargs)
            
def input_with_padding(prompt=''):
    if prompt and not prompt.startswith('  '): 
        padded_prompt = ' ' * 2 + prompt
    else:
        padded_prompt = prompt
    return original_input(padded_prompt)
#Override the built-in print function
builtins.print = print_with_padding
builtins.input = input_with_padding

class Character:
    def __init__(self, name, role, health, courage, strength, intelligence, abilities):
        self.name = name
        self.role = role
        self.health = health
        self.health = 100
        self.max_health = 500
        self.courage = courage
        self.strength = strength
        self.intelligence = intelligence
        self.abilities = abilities
        self.inventory = []
        self.experience = 0
        self.level = 1
        self.rest_count = 0
        self.last_rest_time = 0
        self.rest_cooldown = 240 
        self.max_rests = 3 
    
    def show_stats(self):
        print("\nCharacter Stats:")
        print(f"Name: {self.name}")
        print(f"Role: {self.role}")
        print(f"Health: {self.health}")
        print(f"Courage: {self.courage}")
        print(f"Strength: {self.strength}")
        print(f"Intelligence: {self.intelligence}")
        print(f"Level: {self.level}")
        print(f"XP needed for next level: {(self.level * 100) - self.experience}")
        print(f"Total Experience: {self.experience}")
        print(f"Abilities: {', '.join(self.abilities) if self.abilities else 'None'}")
        print(f"Inventory: {', '.join(self.inventory) if self.inventory else 'Empty'}\n")
    
    def gain_experience(self, amount):
        self.experience += amount
        if self.experience >= self.level * 100:
            self.level_up()
    
    def level_up(self):
        self.level += 1
        self.health += 10
        print(f"{self.name} has leveled up to Level {self.level}!")
        if len(self.abilities) < 5:
            new_ability = f"Ability {len(self.abilities) + 1}"
            self.abilities.append(new_ability)
            print(f"New ability unlocked: {new_ability}")
    
    def can_rest(self, current_time):
        if self.last_rest_time == 0:
            return True 
        #Calculate minutes since last rest and cap the waiting time at 4 hours (240 minutes)
        minutes_since_last_rest = current_time - self.last_rest_time
        if minutes_since_last_rest < 0 or minutes_since_last_rest > 240:
            minutes_since_last_rest = 240
        
        #Allow rest if under 3 rests or if 4 hours have passed
        return self.rest_count < 3 or minutes_since_last_rest >= 240
    
    def rest(self, current_time):
        minutes_since_last_rest = 0 if self.last_rest_time == 0 else current_time - self.last_rest_time
        can_rest = self.last_rest_time == 0 or self.rest_count < 3 or minutes_since_last_rest >= 240
        
        if can_rest:
            if self.last_rest_time == 0 or minutes_since_last_rest >= 240:
                self.rest_count = 0 
            self.health += 30
            self.rest_count += 1
            self.last_rest_time = current_time
            return True, ""
        else:
            remaining_minutes = 240 - minutes_since_last_rest
            remaining_minutes = max(0, remaining_minutes)  #Ensure non-negative
                
            remaining_hours = int(remaining_minutes / 60)
            remaining_mins = int(remaining_minutes % 60)
            
            print(f"\nYou must continue exploring for {remaining_hours} hours and {remaining_mins} minutes before you can rest again.")
            return False
        
    #For artifact collection notification
    def collect_artifact(self, artifact):
        required_artifacts = [
            "Ancient Medallion",
            "Phantom's Locket",
            "Family Secret",
            "Spectral Spice"
        ]
    
        self.inventory.append(artifact)
        print("\n**Important Update!**") 
        print(f"You have gathered the {artifact}, one of the critical artifacts")
        print("required to cleanse the mansion's curse!")
        print("This artifact has been added to your game log.")
    
        #Check if all artifacts have been collected
        collected_required = [item for item in self.inventory if item in required_artifacts]
        if len(collected_required) == len(required_artifacts):
            print("\n**Important Update!**")
            print("You have collected all required artifacts!")
            print("Head to the Secret Chamber at midnight to cleanse the mansion.")
    
    def show_inventory(self):
        if not self.inventory:
            print("\nYour inventory is empty.")
        else:
            print("\nInventory:")
            for item in self.inventory:
                print(f"- {item}")

class Puzzle:
    def __init__(self, name, description, solution, reward):
        self.name = name
        self.description = description
        self.solution = solution
        self.reward = reward
        self.solved = False
        self.attempts = 0
        self.max_attempts = 3
        self.cooldown = False
        self.last_failed_time = 0 
    
    def attempt_solve(self, answer):
        if answer.lower() == self.solution.lower():
            self.solved = True
            return True
        return False
       
PUZZLES = {
    "infinity_room": Puzzle(
        "Infinite Madness Riddle",
        "I am born, but I do not die. I am irrational, yet entirely logical. My existence is everywhere, yet you cannot see me. What am I?",
        ["pi", "Pi"],
        "Ancient Medallion"
    ),
    
    "library": Puzzle(
        "Whispers of the Past",
        "Ancient books line the walls. A mysterious voice echoes: 'In life I was sealed, in death I was revealed. What lies behind my cover tells tales of what time has unveiled.' Which section of books calls to you?",
        "history",
        "Arcane Tome"
    ),
    
    "hidden_study": Puzzle(
        "Alchemist's Riddle",
        "Four bottles before you:\nGreen shows the way,\nRed brings decay,\nBlue tells a lie,\nPurple makes you fly.\nWhich do you drink to reveal the truth?",
        "green",
        "Truth Serum"
    ),
    
    "rotating_wall": Puzzle(
        "The Lunar Phantom",
        "Neither alive nor truly dead, I pull the tides with silent tread.",
        ["moon", "the moon"],
        "Celestial Key"
    ),
    
    "dining_hall": Puzzle(
        "Phantom Feast",
        "Ghostly figures sit at the table. Four place settings remain: First seat has a broken wine glass, second has a shattered plate, third has twisted silverware, fourth is untouched. Which place setting belongs to the living?",
        ["untouched","fourth"],
        "Silver Candlestick"
    ),
    
    "kitchen": Puzzle(
        "Witch's Recipe",
        "An old recipe book lies open. Three bottles are labeled: 'Midnight Dew', 'Wolfsbane', 'Ghost Pepper'. The recipe reads: 'First comes the howling wind, then the spirit's tears, finally the phantom's breath.'",
        "wolfsbane midnight dew ghost pepper",
        "Spectral Spice"
    ),
    
    "childs_bedroom": Puzzle(
        "Music Box Mystery",
        "A music box plays a haunting tune. The inscription reads: 'When darkness falls, three notes play. First comes sorrow, then comes fright, last comes peace in dead of night.' Enter the emotions in order.",
        "sorrow fright peace",
        "Phantom's Locket"
    ),
    
    "guest_room": Puzzle(
        "Portrait Warning",
        "Three portraits hang crooked: Child laughing, Adult weeping, Elder sleeping. Below reads: 'Life's journey shown in tears and breath, which stage comes closest to death?'",
        ["sleeping","elder", "Elder", "elder sleeping"],
        "Family Secret"
    )
}

def initialize_puzzle_cooldowns(puzzles):
        for puzzle_key, puzzle in puzzles.items():
            if not hasattr(puzzle, 'last_failed_time'):
                puzzle.last_failed_time = 0
            if not hasattr(puzzle, 'cooldown_period'):
                puzzle.cooldown_period = 120  #2 hours
initialize_puzzle_cooldowns(PUZZLES)

def create_character():
    print("Choose a character:")
    characters = {
        "1": ("Lost Traveler", 100, 50, 30, 80, ["Survival Instinct"]),
        "2": ("Ghost Hunter", 120, 70, 60, 50, ["Ghost Sense"]),
        "3": ("Escaped Prisoner", 110, 40, 80, 30, ["Adrenaline Boost"]),
        "4": ("Cursed Noble", 90, 90, 40, 70, ["Dark Whisper"]),
        "5": ("Custom Character",)
    }
    
    for key, value in characters.items():
        if key != "5":
            print(f"{key}: {value[0]} (Health: {value[1]}, Courage: {value[2]}, Strength: {value[3]}, Intelligence: {value[4]}, Ability: {value[5][0]})")
        else:
            print(f"{key}: {value[0]}")

    choice = input("Select a number: ")
    
    if choice in characters and choice != "5":
        name = input("Enter your character's name: ")
        role, health, courage, strength, intelligence, abilities = characters[choice]
        return Character(name, role, health, courage, strength, intelligence, abilities)
    
    elif choice == "5":
        name = input("Enter your character's name: ")
        role = input("Enter your character's role: ")
        health = int(input("Enter health (50-150): "))
        courage = int(input("Enter courage (10-100): "))
        strength = int(input("Enter strength (10-100): "))
        intelligence = int(input("Enter intelligence (10-100): "))
        ability = input("Enter your starting ability: ")
        return Character(name, role, health, courage, strength, intelligence, [ability])
    
    else:
        print("Invalid choice. Try again.")
        return create_character()

class TimeSystem:
    def __init__(self):
        self.hour = 20  #Start at 8 PM
        self.minutes = 0
        self.day = 1
        self.midnight_events = False
        
    def advance_time(self, minutes=15):
        self.minutes += minutes
        if self.minutes >= 60:
            self.hour += self.minutes // 60
            self.minutes = self.minutes % 60
            
        if self.hour >= 24:
            self.hour -= 24
            self.day += 1
        
        #Reset midnight events flag when it's no longer midnight
        if self.hour != 0:
            self.midnight_events = False
            
        #Trigger midnight events
        if self.hour == 0 and not self.midnight_events:
            self.trigger_midnight_events()
            self.midnight_events = True
    
    def trigger_midnight_events(self):
        print("\n**Important Update!**")
        print("\nThe mansion's clock strikes midnight...")
        print("You catch the echo of a harpsichord beginning to play")
        print("A cold wind howls through the halls")
        print("The very walls seem to shift and change")
        print("This darkness will certainly spawn mysterious events")
        print("and new places to explore...but for how long?")
        time.sleep(2)
        return {
            "spawn_enemies": True,
            "reveal_passages": True,
            "enhance_difficulty": True
        }
    
    def is_night(self):
        return 20 <= self.hour or self.hour <= 5
    
    def is_midnight(self):
        return self.hour == 0
    
    def get_time_string(self):
        period = "AM" if self.hour < 12 else "PM"
        display_hour = self.hour if self.hour <= 12 else self.hour - 12
        if display_hour == 0:
            display_hour = 12
        return f"Day {self.day}, {display_hour}:{self.minutes:02d} {period}"

class Portal:
    def __init__(self, name, description):
        self.name = name
        self.description = description
        self.destinations = []
    
    def add_destination(self, room_id):
        self.destinations.append(room_id)
    
    def use(self, player, current_room, rooms, time_system):
        print(self.description)
        print("\nAvailable destinations:")
        
        #Create a list of valid destinations (excluding current room)
        valid_destinations = []
        for i, room_id in enumerate(self.destinations):
            if room_id != current_room: 
                valid_destinations.append((i+1, room_id))
                print(f"{i+1}. {rooms[room_id].name}")
        
        print(f"{len(valid_destinations)+1}. Cancel (stay here)")
        
        while True:
            try:
                choice = int(input("\n  Where would you like to go? "))
                if 1 <= choice <= len(valid_destinations):
                    destination = valid_destinations[choice-1][1]
                    print(f"\nThe portal shimmers and glows brightly. You feel a strange sensation as your body dissolves into particles of light...")
                    time_system.advance_time(120) 
                    print(f"Time passes... It is now {time_system.get_time_string()}")
                    return destination
                elif choice == len(valid_destinations)+1:
                    print("You step away from the portal.")
                    return current_room
                else:
                    print("Invalid choice. Please try again.")
            except ValueError:
                print("Please enter a number.")

grand_hall_portal = Portal(
    "Ancient Portal",
    "In the far corner of the Grand Hall, you notice an intricate circle etched into the floor. " +
    "Strange symbols pulse with an ethereal blue light. As you approach, the symbols glow brighter, " +
    "and you feel a gentle pull, as if the circle is inviting you to step inside. " +
    "This appears to be some kind of magical transportation device."
)

dungeon_portal = Portal(
    "Forgotten Escape Rune",
    "Hidden in the shadows of the cell, you discover a faded rune carved into the stone wall. " +
    "As your fingers trace its outline, it begins to emit a soft purple glow. " +
    "The air around it shimmers and distorts, creating what looks like a window to another place. " +
    "Perhaps this was once used by prisoners to escape their confinement."
)

observatory_portal = Portal(
    "Celestial Gateway",
    "The domed ceiling of the Observatory is painted with a map of the night sky. " +
    "As you gaze upward, certain stars begin to twinkle more intensely than others. " +
    "They connect with lines of silver light, forming a constellation unlike any you've seen. " +
    "The floor beneath you reflects this pattern, creating a doorway of starlight. " +
    "This mystical gateway seems to offer passage to distant parts of the mansion."
)

class Room:
    def __init__(self, name, description, can_rest=False, night_description=None, midnight_only=False, puzzle=None):
        self.name = name
        self.description = description
        self.can_rest = can_rest
        self.explored = False
        self.enemies = []
        self.items = []
        self.exits = {}
        self.night_description = night_description if night_description else description
        self.midnight_only = midnight_only
        self.locked = False
        self.night_only = False
        self.requires_item = None
        self.puzzle = puzzle
        self.npcs = []
        self.night_npcs = []
        self.items = []
        self.potions = []
        self.portal = None
        
    def add_puzzle(self, puzzle):
        self.puzzle = puzzle
    
    def add_potion(self, potion):
        self.potions.append(potion)
    
    def add_enemy(self, enemy_type):
        self.enemies.append(enemy_type)
    
    def add_item(self, item):
        self.items.append(item)
    
    def add_exit(self, direction, room_id):
        self.exits[direction] = room_id
    
    def get_description(self, time_system):
        if self.name == "Grand Hall":
            if time_system.is_night():
                return "A magnificent hall with high ceilings and ancient tapestries. Moonlight streams through tall windows, casting eerie shadows across the marble floor."
            else:
                return "A magnificent hall with high ceilings and ancient tapestries. Sunlight pours through tall windows, illuminating specks of dust dancing in the air."
        if time_system.is_night():
            return self.night_description
        else:
            return self.description
    
    def enter(self, player, combat_system, time_system):
        if self.midnight_only and not is_night_time(time_system):
            print("This room is only accessible during the night (midnight to 5am).")
            return False
        if not self.explored:
            print(f"\nYou enter {self.name}.")
            print(self.get_description(time_system))
            player.gain_experience(20)
            self.explored = True
        else:
            print(f"\nYou return to {self.name}.")
        
        if self.enemies:
            enemy = random.choice(self.enemies)
            if not combat_system.initiate_combat(player, enemy):
                print("Game Over!")
                sys.exit()
            self.enemies.remove(enemy)
        
        if self.items:
            print(f"\nYou find: {', '.join(self.items)}")
            take_items = input("Take items? (yes/no): ").lower()
            if take_items in ["yes", "y"]:
                player.inventory.extend(self.items)
                self.items = []
        
        #Only prompt for rest once upon entering a room that allows it
        if self.can_rest:
            current_time = time_system.day * 24 * 60 + time_system.hour * 60 + time_system.minutes
            #Check if player can rest before prompting
            minutes_since_last_rest = 0 if player.last_rest_time == 0 else current_time - player.last_rest_time
            can_rest_check = player.last_rest_time == 0 or player.rest_count < 3 or minutes_since_last_rest >= 240
            
            if can_rest_check:
                rest_prompt = input("Would you like to rest here? (yes/no): ").lower()
                if rest_prompt in ['yes', 'y']:
                    rest_result = player.rest(current_time)
                    if isinstance(rest_result, tuple):
                        can_rest, rest_message = rest_result
                    else:
                        can_rest = rest_result
                        rest_message = ""
                    
                    if can_rest:
                        # Rest was successful, advance time
                        time_system.advance_time(60)  # Advance by 60 minutes (1 hour)
                        print(f"{player.name} has rested and recovered health. Current health: {player.health}")
                        print(f"\nTime passes... It is now {time_system.get_time_string()}")
        return True
    
class CombatSystem:
    def __init__(self):
        self.enemy_types = {
            "Ghost": {"health": 50, "damage": 10, "xp": 30},
            "Skeleton": {"health": 40, "damage": 15, "xp": 25},
            "Possessed Armor": {"health": 80, "damage": 20, "xp": 45},
            "Shadow": {"health": 30, "damage": 25, "xp": 35},
            "Wraith": {"health": 70, "damage": 15, "xp": 40},
            "Poltergeist": {"health": 45, "damage": 20, "xp": 45}
        }
        self.midnight_enemies = {
            "Demon": {"health": 100, "damage": 30, "xp": 60},
            "Nightmare": {"health": 80, "damage": 35, "xp": 55}
        }
    
    def initiate_combat(self, player, enemy_type):
        enemy = self.enemy_types[enemy_type].copy()
        print(f"\nA {enemy_type} appears!")
        
        while enemy["health"] > 0 and player.health > 0:
            print(f"\nYour Health: {player.health} | Enemy Health: {enemy['health']}")
            action = self.combat_menu(player)
            
            if action == "attack":
                damage = max(5, player.strength // 3 + random.randint(1, 10))
                enemy["health"] -= damage
                print(f"You deal {damage} damage to the {enemy_type}!")
            
            elif action == "ability" and player.abilities:
                damage = max(10, player.intelligence // 2 + random.randint(5, 15))
                enemy["health"] -= damage
                print(f"You use {player.abilities[0]} and deal {damage} damage!")
            
            if enemy["health"] > 0:
                player.health -= enemy["damage"]
                print(f"The {enemy_type} attacks you for {enemy['damage']} damage!")
        
        if player.health > 0: #Victory! 
            xp_gained = random.randint(20, 30)
            player.experience += xp_gained
            print(f"\nCongratulations! You defeated the {enemy_type}!")
            print(f"Experience gained: {xp_gained} XP")
            
            #Calculate XP needed for next level (100 XP per level)
            xp_for_next = (player.level * 100) - player.experience
            print(f"XP needed for next level: {xp_for_next}")
            #Check for level up
            if player.experience >= player.level * 100:
                player.level_up()
            player.gain_experience(enemy["xp"])
            print(f"Total XP: {player.experience}")
            return True
        return False
    
    def combat_menu(self, player):
        print("\nCombat Options:")
        print("1. Attack")
        print("2. Use Ability")
        #Dynamically show all earned abilities
        for i, ability in enumerate(player.abilities, start=3):
            print(f"{i}. {ability}")
        #print("3. Attempt to Flee")
        choice = input("Choose your action: ")
        return "attack" if choice == "1" else "ability"
    
def use_potion(player):
    if not any(item.startswith("Health Potion") for item in player.inventory):
        print("You don't have any potions!")
        return
    
    potions = [item for item in player.inventory if item.startswith("Health Potion")]
    print("\nAvailable Potions:")
    for i, potion in enumerate(potions, 1):
        print(f"{i}. {potion}")
    
    choice = input("  Use which potion? (number or 'cancel'): ")
    if choice.lower() == 'cancel':
        return
    
    try:
        index = int(choice) - 1
        potion = potions[index]
        restore_amount = 50 if potion == "Health Potion" else 100
        player.health = min(player.health + restore_amount, 150)
        player.inventory.remove(potion)
        print(f"Used {potion}. Health restored to {player.health}")
    except:
        print("Invalid choice!")

def solve_puzzle(puzzle, player, time_system=None):
    if puzzle.solved:
        print("\nYou've already solved this puzzle.")
        return True
    
    #If time_system is not provided, we can't implement the cooldown
    if time_system is None:
        print("\nWarning: Time system not available, cooldown cannot be enforced.")
        #Proceed with simplified puzzle solving without cooldown
        return False
    
    current_time = time_system.day * 24 * 60 + time_system.hour * 60 + time_system.minutes
    
    if puzzle.attempts >= 3 and current_time - puzzle.last_failed_time < puzzle.cooldown_period:
        #Calculate remaining cooldown time
        remaining_minutes = puzzle.cooldown_period - (current_time - puzzle.last_failed_time)
        remaining_hours = remaining_minutes // 60
        remaining_mins = remaining_minutes % 60
        print("\nYou should take some time to think about this puzzle.")
        print("Come back later with a fresh perspective.")
        return False
    
    if puzzle.attempts >= 3 and current_time - puzzle.last_failed_time >= puzzle.cooldown_period:
        puzzle.attempts = 0
        
    print(f"\n{puzzle.description}")
    answer = input("\n  Enter your solution: ").lower().strip()
    
    #Check for empty input
    if not answer:  
        print("\nYou must provide an answer to solve the puzzle.")
        return False
         
    if isinstance(puzzle.solution, list):
        if answer in puzzle.solution:
            print(f"\nCorrect! You've solved the {puzzle.name}!")
            print(f"You receive: {puzzle.reward}")
            required_artifacts = [
                "Ancient Medallion",
                "Phantom's Locket",
                "Family Secret",
                "Spectral Spice"
            ]
        
            if puzzle.reward in ["Ancient Medallion", "Phantom's Locket", "Family Secret", "Spectral Spice"]:
                player.collect_artifact(puzzle.reward)
            else:
                player.inventory.append(puzzle.reward)
            
            puzzle.solved = True
            return True
    else:
        if answer == puzzle.solution.lower():
            print(f"\nCorrect! You've solved the {puzzle.name}!")
            print(f"You receive: {puzzle.reward}")
            required_artifacts = [
                "Ancient Medallion",
                "Phantom's Locket",
                "Family Secret",
                "Spectral Spice"
            ]
        
        if puzzle.reward in ["Ancient Medallion", "Phantom's Locket", "Family Secret", "Spectral Spice"]:
            player.collect_artifact(puzzle.reward)
        else:
            player.inventory.append(puzzle.reward)
            
        puzzle.solved = True
        return True
    
    puzzle.attempts += 1
    if puzzle.attempts >= 3:
        #Set the last failed time to current time
        puzzle.last_failed_time = current_time
        print("\nYou've made too many incorrect attempts. Take some time to think about the puzzle.")
        print("Come back in 2 hours (game time) to try again.")
    else:
        print(f"\nIncorrect. You have {3 - puzzle.attempts} attempts remaining.")
    return False

def search_master_bedroom(player):
    print("\nYou carefully examine the master bedroom...")
    print("The ambient light reveals the details of the dusty furniture.")
    print("Your attention is drawn to an ornate bookcase against the far wall.")
    
    choice = input("\n  Would you like to inspect the bookcase more closely? (yes/no): ").lower()
    
    if choice in ["yes", "y"]:
        print("\nAs you run your fingers along the ancient tomes, one leather-bound book")
        print("feels different from the others. It's slightly warm to the touch.")
        
        choice = input("\n  Attempt to pull the book? (yes/no): ").lower()
        
        if choice in ["yes", "y"]:
            print("\n*click*")
            print("The book tilts backward, and with a grinding sound,")
            print("the entire bookcase slides smoothly to the side.")
            print("\nBehind the wall, you discover a small alcove filled with arcane items.")
            print("Two vials catch your eye:")
            print("- A sinister purple vial labeled 'Essence of Shadow'")
            print("- A gleaming golden potion marked 'Vitality Elixir'")
            
            if "Essence of Shadow" not in player.inventory and "Vitality Elixir" not in player.inventory:
                player.inventory.extend(["Essence of Shadow", "Vitality Elixir"])
                print("\nYou carefully add both vials to your inventory.")
                print("The Essence of Shadow can be used to deal devastating damage to enemies.")
                print("The Vitality Elixir will restore your health in times of need.")
            else:
                print("\nYou've already collected the items from this secret alcove.")

def setup_room_connections(rooms):
    #Ground Floor
    rooms["foyer"].add_exit("north", "dining_hall")
    rooms["foyer"].add_exit("east", "library")
    rooms["foyer"].add_exit("stairs", "upper_hall")
    
    #Main Floor Connections
    rooms["dining_hall"].add_exit("south", "foyer")
    rooms["dining_hall"].add_exit("east", "kitchen")
    rooms["dining_hall"].add_exit("west", "grand_hall")
    
    #Grand Hall connections
    rooms["grand_hall"].add_exit("east", "dining_hall")

    #Kitchen connections
    rooms["kitchen"].add_exit("west", "dining_hall")
    rooms["dining_hall"].add_exit("east", "kitchen")
    
    #Library Wing
    rooms["library"].add_exit("west", "foyer")
    rooms["library"].add_exit("rotate bookshelf", "hidden_study")
    
    #Secret Areas
    rooms["hidden_study"].add_exit("rotate back", "library")
    rooms["hidden_study"].add_exit("trapdoor", "dungeon")
    
    #Dungeon System
    rooms["dungeon"].add_exit("tunnel", "secret_tunnel")
    rooms["secret_tunnel"].add_exit("up", "foyer")
    rooms["dungeon"].add_exit("north", "dungeon_cell")
    rooms["dungeon_cell"].add_exit("south", "dungeon")
    
    #Upper Floor
    rooms["upper_hall"].add_exit("down", "foyer")
    rooms["upper_hall"].add_exit("east to bedroom 1", "master_bedroom")
    rooms["upper_hall"].add_exit("east to bedroom 2", "childs_bedroom")
    rooms["upper_hall"].add_exit("west", "guest_room")
    rooms["upper_hall"].add_exit("north", "infinity_room")
    
    #Observatory
    rooms["observatory"].add_exit("down", "upper_hall")
    
    #Master Bedroom connections
    rooms["master_bedroom"].add_exit("west", "upper_hall")
    
    #Child's Room connections
    rooms["childs_bedroom"].add_exit("west", "upper_hall")
    
    #Guest Room connections
    rooms["guest_room"].add_exit("east", "upper_hall")
    
    #Hidden Passages
    rooms["infinity_room"].add_exit("rotate wall", "secret_chamber")
    rooms["infinity_room"].add_exit("south", "upper_hall")
    rooms["secret_chamber"].add_exit("passage", "library")
    
    #Add portals to rooms
    rooms["grand_hall"].portal = grand_hall_portal
    rooms["dungeon_cell"].portal = dungeon_portal
    rooms["observatory"].portal = observatory_portal
    
    #Add destinations to each portal
    grand_hall_portal.add_destination("dungeon_cell")
    grand_hall_portal.add_destination("observatory")
    grand_hall_portal.add_destination("library")

    dungeon_portal.add_destination("grand_hall")
    dungeon_portal.add_destination("observatory")
    dungeon_portal.add_destination("kitchen")

    observatory_portal.add_destination("grand_hall")
    observatory_portal.add_destination("dungeon_cell")
    observatory_portal.add_destination("master_bedroom")

def process_player_choice(choice, player, rooms, current_room, time_system, combat_system):
    if choice == "1":
        print("\nAvailable exits:")
        for direction, room_id in rooms[current_room].exits.items():
            if direction == "rotate bookshelf" and not is_night_time(time_system):
                continue
            print(f"- {direction}: {rooms[room_id].name}")
        
        direction = input("\n  Which direction? ").lower()
        if direction in rooms[current_room].exits:
            if direction == "rotate bookshelf" and not is_night_time(time_system):
                print("The bookshelf seems immovable at this time. Perhaps try again during the night (midnight to 5am)...")
                return current_room
            current_room = rooms[current_room].exits[direction]
            if 6 <= time_system.hour < 20:  #Between 6AM and 8PM
                time_system.advance_time(60)  #Advance by 1 hour during day
            else:
                time_system.advance_time(30)  #Advance by 30 minutes during night
    
    elif choice == "2":
        if rooms[current_room].can_rest:
            current_time = time_system.day * 24 * 60 + time_system.hour * 60 + time_system.minutes
            rest_result = player.rest(current_time)
            
            #Handle both possible return types (boolean or tuple)
            if isinstance(rest_result, tuple):
                can_rest, rest_message = rest_result
            else:
                can_rest = rest_result
                rest_message = ""
            
            if can_rest:
                time_system.advance_time(60) 
                print(f"{player.name} has rested and recovered health. Current health: {player.health}")
                print(f"\nTime passes... It is now {time_system.get_time_string()}")
                
                #Random encounter chance after resting
                if random.random() < 0.3:
                    combat_system.initiate_combat(player, "Shadow")
            elif rest_message:  
                #Player couldn't rest, show the message
                print(rest_message)
            else:
                print("You have rested too much recently.")
        else:
            print("You cannot rest here.")
    
    elif choice == "3":
        player.show_stats()
        
        print(rooms[current_room].get_description(time_system))
        #Check if there's a portal in current room
        if rooms[current_room].portal:
            print("\nYou notice something unusual in this room...")
            print(f"There appears to be a {rooms[current_room].portal.name} here.")
            interact = input("Would you like to examine it closer? (yes/no): ").lower()
            if interact in ["yes", "y"]:
                #Use the portal
                new_room = rooms[current_room].portal.use(player, current_room, rooms, time_system)
                if new_room != current_room:
                    #Player teleported
                    current_room = new_room
                    rooms[current_room].enter(player, combat_system, time_system)
    
    elif choice == "4":
        player.show_inventory()
    
    elif choice == "5":
        check_game_log(player)
    
    elif choice == "6":
        use_potion(player)
    
    elif choice == "7" and rooms[current_room].puzzle and not rooms[current_room].puzzle.solved:
        solve_puzzle(rooms[current_room].puzzle, player, time_system)
    
    elif choice == "7" and current_room == "master_bedroom":
        search_master_bedroom(player)
    
    elif choice == "7" and rooms[current_room].portal:
        new_room = rooms[current_room].portal.use(player, current_room, rooms, time_system)
        if new_room != current_room:
            current_room = new_room
            rooms[current_room].enter(player, combat_system, time_system)
    
    return current_room

#Helper function to check if it's night time 
def is_night_time(time_system):
    return 0 <= time_system.hour < 5

def is_victory_time(time_system):
    return 0 <= time_system.hour < 2  #Midnight to 2AM window for victory condition

def use_item(player, combat_mode=False):
    if not player.inventory:
        print("\nYour inventory is empty.")
        return
    
    print("\nAvailable items:")
    for idx, item in enumerate(player.inventory, 1):
        print(f"{idx}. {item}")
    
    choice = input("\n  Which item would you like to use? (or 'cancel'): ")
    if choice.lower() == 'cancel':
        return
    
    try:
        index = int(choice) - 1
        item = player.inventory[index]
        
        if combat_mode:
            if "Essence of Shadow" in item:
                print("\nYou unleash the dark essence upon your enemy!")
                player.inventory.remove(item)
                return 50
        else:
            if any(healing_item in item for healing_item in ["Elixir", "Potion", "Vitality"]):
            #if "Elixir" in item or "Potion" in item:
                print("\nYou drink the healing potion.")
                player.health += 50
                player.inventory.remove(item)
                print(f"Health restored! Current health: {player.health}")
    except (ValueError, IndexError):
        print("\nInvalid selection.")

def check_game_log(player):
    print("\nGame Objectives:")
    print("---------------")
    print("Required Artifacts to Cleanse the Mansion:")
    required_artifacts = [
        "Ancient Medallion",
        "Phantom's Locket",
        "Family Secret",
        "Spectral Spice"
    ]
    
    for artifact in required_artifacts:
        status = "[Acquired]" if artifact in player.inventory else "[Still need]"
        print(f"{artifact} {status}")
    
    print("\nFinal Objective:")
    print("Bring all artifacts to the Secret Chamber between midnight and 2AM")
    print("\nCurrent Progress:")
    print(f"Artifacts collected: {sum(1 for item in required_artifacts if item in player.inventory)}/{len(required_artifacts)}")

def check_victory_conditions(player, current_room, time_system):
    required_artifacts = ["Ancient Medallion", "Phantom's Locket", "Family Secret", "Spectral Spice"]
    
    if all(item in player.inventory for item in required_artifacts):
        if current_room == "secret_chamber" and is_victory_time(time_system):
            print(f"\nAs {player.name} enters the secret chamber,")
            print("\nThe artifacts resonate with ancient power!")
            print("The mansion's dark energy dissipates as the items glow brilliantly.")
            print("You've successfully cleansed the mansion of its evil presence.")
            print(f"\nCongratulations, {player.name}! You've won the game!")
            print(f"Final Score: Level {player.level}, Experience: {player.experience}")
            input("\n  Press Enter to exit...")
            sys.exit()

def main():
    print("\n" + "Welcome to Zork: The Haunted Mansion Adventure RPG!".center(terminal_width)),
    print("Created & Developed by Chad Jordan using Python 2.5".center(terminal_width)), 
    print("and updated for Python 3".center(terminal_width)),
    print("")
    player = create_character()
    combat_system = CombatSystem()
    time_system = TimeSystem()
    
    rooms = {
        "foyer": Room("The Foyer", "A grand entrance with cobweb-covered chandeliers and eerie silence.",
                     night_description="The chandelier casts dancing shadows in the moonlight, and whispers of suffering echo through the halls."),
        "library": Room("The Library", "Bookshelves line the walls, filled with ancient texts and manuscripts.", can_rest=True),
        "hidden_study": Room("Hidden Study", "A secret room filled with forbidden knowledge.", midnight_only=True, puzzle=PUZZLES["hidden_study"]),
        "childs_bedroom": Room("A Dusty Child's Bedroom", "Something terrible happened in here. The faint echoes of crying children fills the room.", puzzle=PUZZLES["childs_bedroom"]),
        "grand_hall": Room("Grand Hall", "A magnificent hall with high ceilings and ancient tapestries."),
        "dining_hall": Room("The Dining Hall", "A long table set with dusty plates and rotting food.", puzzle=PUZZLES["dining_hall"]),
        "kitchen": Room("The Kitchen", "Rusty pots and pans hang from the ceiling, but sway and clatter together from an unknown evil force.", puzzle=PUZZLES["kitchen"]),
        "dungeon": Room("Dungeon", "A damp, dark and cold place with torches aligning the walls filled with skeletal remains."),
        "secret_tunnel": Room("Secret Tunnel", "A narrow passage carved through stone."),
        "upper_hall": Room("Upper Hall", "A long, darkened corridor with watching portraits. Likely the remnants of the forgotten family who lived here."),
        "master_bedroom": Room("Master Bedroom", "A luxurious bedroom with a four-poster bed.", can_rest=True),
        "infinity_room": Room("Infinity Room", "Walls of mirrors reflect infinite possibilities. This room looks to go on forever..."),
        "secret_chamber": Room("Secret Chamber", "A hidden room that few have discovered."),
        "guest_room": Room("Guest Room", "A dusty bedroom that once welcomed visitors. The air feels heavy with untold stories", can_rest=True, puzzle=PUZZLES["guest_room"]),
        "observatory": Room("Observatory", "A circular room topped with a glass dome. Astronomical instruments of brass and silver are arranged around the room, including a large telescope pointed at the celestial sky.", 
                   night_description="The Observatory is bathed in moonlight streaming through the glass dome, and eerie shadows are cast from the moonlight."),
        "dungeon_cell": Room("Dungeon Cell", "A cold, lonely cell with stone walls and rusted iron bars.",
                    night_description="In the darkness, the dungeon cell casts eerie shadows from the moonlight through a small window.")

    }
    
    #Add enemies and items to rooms
    rooms["library"].add_enemy("Ghost")
    rooms["library"].add_item("Ancient Tome")
    rooms["childs_bedroom"].add_enemy("Shadow")
    rooms["childs_bedroom"].add_item("Silver Key")
    rooms["dining_hall"].add_enemy("Possessed Armor")
    rooms["kitchen"].add_enemy("Poltergeist")
    rooms["kitchen"].add_item("Rusty Knife")
    rooms["dungeon"].add_enemy("Skeleton")
    
    #Set up room connections
    setup_room_connections(rooms)
    
    #Add puzzles to rooms
    rooms["infinity_room"].add_puzzle(PUZZLES["infinity_room"])
    rooms["library"].add_puzzle(PUZZLES["library"])
    rooms["hidden_study"].add_puzzle(PUZZLES["hidden_study"])
    rooms["upper_hall"].add_puzzle(PUZZLES["rotating_wall"])
    
    #Add items and potions
    rooms["hidden_study"].items.append("Mysterious Scroll")
    rooms["library"].potions.append({"name": "Health Potion", "restore": 50})
    rooms["master_bedroom"].potions.append({"name": "Greater Health Potion", "restore": 100})
    
    current_room = "foyer"
    
    #Main game loop
    while True:
        print(f"\n{time_system.get_time_string()}")
        
        if rooms[current_room].midnight_only and not is_night_time(time_system):
            print("This room is only accessible during the night (midnight to 5am).")
            #Move player back to a valid room
            current_room = "library" 
            continue
            
        rooms[current_room].enter(player, combat_system, time_system)
        
        print("\nActions:")
        print("1. Move to another room")
        print("2. Rest")
        print("3. View player stats")
        print("4. View inventory")
        print("5. View game log")
        print("6. Use potion")
        if rooms[current_room].puzzle and not rooms[current_room].puzzle.solved:
            print("7. Solve puzzle")
        if current_room == "master_bedroom":
            print("7. Search the room")
        if rooms[current_room].portal:
            print(f"7. Use the {rooms[current_room].portal.name}")
        print("Q - Quit Game")
        
        choice = input("\n  What would you like to do? ")
        if choice in ['q', 'Q', 'quit', 'exit']:
            confirm = input("\n  Are you sure you wish to depart the game? (yes/no): ").lower()
            if confirm in ['yes', 'y']:
                print("\nFarewell, brave adventurer! The mansion's mysteries will await your return...")
                time.sleep(2)
                sys.exit()
            else:
                continue
            
        current_room = process_player_choice(choice, player, rooms, current_room, time_system, combat_system)

        if check_victory_conditions(player, current_room, time_system):
            print(f"\nAs {player.name} enters the secret chamber,")
            print("\nthe artifacts resonate with ancient power!")
            print("The mansion's dark energy dissipates as the items glow brilliantly.")
            print("You've successfully cleansed the mansion of its evil presence.")
            print(f"\nCongratulations, {player.name}! You've won the game!")
            print(f"Final Score: Level {player.level}, Experience: {player.experience}")
            break
        
if __name__ == "__main__":
    main()